# MLforecasting

Framework for testing and development of ML prediction models

# How to launch MLaaS framework

Please start  by cloning the repository into your PC.

0. **Extra**
    Before initiating the framework, an extra step is needed.
    First, go to file clientsML_forecast/conductor/config.json, and in the section "volume", change "-<abs_path>" to the full path locally, allowing the code to setup the volumes properly.
    Secondly, you will also need to change the ip address of the model container's API. To do so, check the ip address in line 183 of conductor.py.
    Run ```ip a``` on Ubuntu to check your machine's IP adress and change it to the correct address. 

1. If docker images of the ML models and/or conductor do not yet exist, create them.
    To create the ML model images, from root directory, run the following: ```docker build --build-arg ARCH=amd64 -t <EXAMPLE>_ml -f clientsML_forecast/<EXAMPLE>/Dockerfile.clientML_AND_interface .```
        Example: 
        ```
        docker build --build-arg ARCH=amd64 -t cnn_ml -f clientsML_forecast/CNN/Dockerfile.clientML_AND_interface .
        ```
    To create the conductor image, run the following commands, starting from base directory:
    ```
        cd clientsML_forecast/conductor
        docker build -t conductor .
    ```

2. To launch the controller, run the following command from base directory:
    ```
    docker-compose -f base-docker-compose.yaml up -d
    ```
    This will launch the conductor model, which will listen for requests in port 8000
    


3. To perform a request to the framework, use the following structure:
```
    curl -X POST http://localhost:8000/data -H "Content-Type: application/json" -d '{"MLmodel": "cnn", "Dataset": "nvehicles_by_datetime_light_August_2021.csv"}'

```

    NOTE: For now, it is recommended that  the values for MLmodel are one of the following: **'lstm', 'cnn', 'rnn', 'arima', 'rf', 'xgb', 'svm', 'sarima'**

4. The framework will then perform predictions using the given model and return the set of predictions, as well as a graph comparing this predictions to the real values.
    This information will be available inside the folder /logs (searching from base directory), in a folder named after the date and time the prediction was made.


# Components

1. ./common - Contains some useful functions that are used both in the interface and on some models.
2. ./interface - Contains the FastAPI service that receives the requests from the controller calls the relevant model functions
3. ./clientsML_forecast - Contains the folders with the building blocks for each model, as well as the controller and preset datasets
4. ./logs (only created after starting the framework) -  Wil receive the folders with model outputs.

# Model image building structure

- **entrypoint.sh** - a startup script to set up the runtime environment for a machine learning application
- **requirementsML.txt** - a file containing all the libraries that need to be installed for the specific forecasting model to be used
- **clientML.py** - a python script that contains all the code needed to preprocess data, train the model, make predictions and get the needed outputs.
- **Dockerfile** - which gathers the other files in the folder as well as the dataset to build an image to be tested.

# Conductor building structure
- **main.py** - creates separate threads for entryserver and conductor programs
- **entryserver.py** - Receives the instructions from user and writes them to a config file.
- **conductor.py** - Reads config file when changes are made, and create the container for model testing
- **config.json** - Stores desired configuration of the model to test


# FAQ
If the framework is instructed to test the same model and dataset two consecutive times, it will not execute the second time.
If you want to do it, just request a different pairing in between.