import os
import time
import re
from datetime import timedelta
from datetime import date, datetime
from sklearn.model_selection import KFold
import matplotlib.pyplot as plt
import pandas as pd
import tensorflow as tf
from keras.callbacks import EarlyStopping
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score, explained_variance_score
from sklearn.preprocessing import MinMaxScaler

from common import logger
from common import containers_info

from scipy import stats
import numpy as np
global dataset
dataset = os.environ.get("DATASET")
logging_level = logger.getLevelName(os.environ.get("LOGGING_LEVEL"))

x_train = []
y_train = []
x_test = []
y_test = []
scaler = None
global_filename = ""
global_date = ""
occupied = 0

# # [] -> actual parameters, 0 -> value of the metric
# parameters_best = ([],0)
# parameters_before_last_agg = ([], 0)

def get_parameters():
    log.info("Getting Parameters!")
    global model
    parameters = model.get_weights()

    return parameters


def set_parameters(parameters):
    global model
    model.set_weights(parameters)
    log.info("Parameters Set!")


def in_usage(value):
    global occupied
    occupied = value
    if value == 1:
        log.info("ML app now in use, blocked")
    elif value == 0:
        log.info("ML app is now freed")




model = tf.keras.models.Sequential([
    tf.keras.layers.InputLayer(shape=(3, 1)),
    
    # Conv1D with L2 regularization and BatchNormalization
    tf.keras.layers.Conv1D(filters=24, kernel_size=2, activation="relu", kernel_regularizer=tf.keras.regularizers.l2(0.01)),
    tf.keras.layers.BatchNormalization(),
    
    tf.keras.layers.Flatten(),
    tf.keras.layers.Dropout(0.05),  # Dropout layer to prevent overfitting
    tf.keras.layers.Dense(1)
])

model.compile(
    optimizer=tf.keras.optimizers.Adam(learning_rate=1e-3), 
    loss=tf.keras.losses.MeanAbsoluteError(), 
    metrics=[tf.keras.metrics.MeanSquaredError()]
)


# Configure logging
log = logger.setup_logger(containers_info.get_current_container_name(), logging_level)


def split_data(df,split_index):
    global x_train, y_train, x_test, y_test
    
    # Validate the split index
    if split_index >= len(df):
        log.info(split_index)
        log.info(len(df))
        raise ValueError(f"Split index {split_index} is greater than or equal to the length of the DataFrame {len(df)}.")
    split_index = int(df.shape[0]*0.8)
    # Select training data
    x_train = df.iloc[:split_index][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_train = df.iloc[:split_index][["car_sum_scaler"]]
    
    # Select testing data
    x_test = df.iloc[split_index:][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_test = df.iloc[split_index:][["car_sum_scaler"]]
    
    # Print the shapes of the resulting datasets (optional)
    print("Shapes:")
    print("x_train:", x_train.shape)
    print("y_train:", y_train.shape)
    print("x_test:", x_test.shape)
    print("y_test:", y_test.shape)



def get_data():
    ##########################################################################################
    # ASK TO ANOTHER ENTITY WICH IS THE DATASET FILE AND OTHER RESTRICTIONS
    file_name = dataset
    date = "2024-05-16"
    ##########################################################################################

    global global_filename
    global_filename = str(file_name).split(".")[0]
    global global_date
    global_date = str(date)

    df = pd.read_csv("/app/dataset/" + str(file_name), index_col="timestamp", parse_dates=True)
    day = pd.to_datetime(str(date))
    #df = df[day - timedelta(days=28) :]

    start_dataset = df.index[0]
    end_dataset = df.index[-1]
    log.info("start timestamp: " + str(start_dataset))
    log.info("end timestamp: " + str(end_dataset))

    global scaler
    scaler = MinMaxScaler()
    scaler.fit(df["car_sum"].to_numpy().reshape(-1, 1))
    df["car_sum_scaler"] = scaler.transform(df["car_sum"].to_numpy().reshape(-1, 1))
    log.info(df["car_sum_scaler"])
    log.info(len(df))
    # Calculate the split index
    split_index = int(len(df)*0.20) 

    df["car_sumB"] = df["car_sum"].shift((split_index))
    df["car_sumBB"] = df["car_sum"].shift((split_index * 2))
    df["car_sumBBB"] = df["car_sum"].shift((split_index * 3))
   
    df["car_sumB_scaler"] = df["car_sum_scaler"].shift((split_index))
    df["car_sumBB_scaler"] = df["car_sum_scaler"].shift((split_index * 2))
    df["car_sumBBB_scaler"] = df["car_sum_scaler"].shift((split_index * 3))
    log.info(df["car_sumB_scaler"].dropna())
    log.info(df["car_sumBB_scaler"].dropna())
    log.info(df["car_sumBBB_scaler"].dropna())
    
    df["car_sum_diff"] = df["car_sum"].diff()  # Calculate difference between consecutive values
    log.info(df.head())
    log.info(len(df))
    df.fillna(0, inplace=True)  # Example
    log.info(df.head())
    log.info(len(df))
    # Call the function to split the data
    split_data(df, split_index)


    # parameters_current = get_parameters()
    # metric_value_current = evaluate()[2]
    # global parameters_before_last_agg
    # global parameters_best
    # parameters_before_last_agg = (parameters_current, metric_value_current)
    # parameters_best = (parameters_current, metric_value_current)

    log.info("Get data successfully!")

    return len(x_train)


def fit(config):
    log.info("----------------  FIT  ----------------- ")
    log.debug("config: " + str(config))

    global model
    global x_train
    global y_train
    global x_test
    global y_test

    es = EarlyStopping(monitor="val_loss", mode="min", patience=10)
  
    # Learning Rate Scheduler Callback
    lr_scheduler = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=5, min_lr=1e-6)
    log.info(f"x_train len {len(x_train)}")

    k = 5
    kf = KFold(n_splits=k, shuffle=True, random_state=42)
    results = []
    start_time_fit = time.time()
    for fold, (train_index, val_index) in enumerate(kf.split(x_train)):
        log.info(f"Training fold {fold + 1}...")

        es = tf.keras.callbacks.EarlyStopping(monitor="val_loss", mode="min", patience=10)
        lr_scheduler = tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=5, min_lr=1e-6)
        
        history = model.fit(
            x_train, y_train,
            epochs=int(config["epochs"]),
            batch_size=int(config["batch_size"]),
             validation_data=(x_test, y_test),
            callbacks=[es, lr_scheduler],
            verbose=0
        )
        
        val_predictions = model.predict(x_test)
        mae = mean_absolute_error(y_test, val_predictions)
        results.append(mae)
        log.info(f"Fold {fold + 1} MAE: {mae}")

    avg_mae = np.mean(results)
    log.info(f"Average MAE over {k} folds: {avg_mae}")

    log.info("Fit Duration: " + str((time.time() - start_time_fit)))

    results = {
        "loss": history.history["loss"][0],
        # "r2_score": history.history["r2_score"][0],
        # "val_loss": history.history["val_loss"][0],
        # "val_r2_score": history.history["val_r2_score"][0],
    }

    log.info("Fit Results: " + str(results))

    return len(x_train), results


def evaluate(config):
    log.info("----------------  EVALUATE  ----------------- ")
    log.debug("config: " + str(config))

    global model
    global x_test
    global y_test

    start_time_evaluate = time.time()
    loss, r_square = model.evaluate(x_test, y_test, steps=int(config["val_steps"]))
    log.info("Evaluate Duration: " + str((time.time() - start_time_evaluate)))

    return float(loss), len(x_test), {"r2_score": float(r_square)}


def predict(plot_graphs, before_after):
    global model
    global scaler
    global x_test
    global y_test

    # Make predictions
    result_scaler = model.predict(x_test)
    result = scaler.inverse_transform(result_scaler)

    car_sum_real = scaler.inverse_transform(y_test)

    # Prepare the result DataFrame
    df_result = x_test.drop(columns=["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"])
    df_result["car_sum_real_scaler"] = y_test
    df_result["result_scaler"] = result_scaler
    df_result["car_sum_real"] = car_sum_real
    df_result["result"] = result

    # Calculate metrics
    mse = mean_squared_error(df_result["car_sum_real"], df_result["result"])
    rmse = np.sqrt(mse)
    mae = mean_absolute_error(df_result["car_sum_real"], df_result["result"])
    r2 = r2_score(df_result["car_sum_real"], df_result["result"])
    explained_variance = explained_variance_score(df_result["car_sum_real"], df_result["result"])

    # Create today's path based on the current date
    today = datetime.now()
    today_path = f"/app/logs/CNN_{dataset}_{today}"
    if not os.path.exists(today_path):
        os.makedirs(today_path)

    # Define the log file path
    log_file = os.path.join(today_path, 'metrics_log.txt')

    # Write metrics to the file
    with open(log_file, 'a') as f:
        f.write(f"Dataset: {global_filename}")
        f.write(f"Evaluation MSE: {mse:.3f}\n")
        f.write(f"Evaluation RMSE: {rmse:.3f}\n")
        f.write(f"Evaluation MAE: {mae:.3f}\n")
        f.write(f"Evaluation R2_SCORE: {r2:.3f}\n")
        f.write(f"Evaluation Explained Variance: {explained_variance:.3f}\n")
        f.write("\n")  # Add a newline for separation between log entries

    if plot_graphs != "false":
        plot_graphs_fn(before_after, str(global_filename), str(global_date), df_result, today, today_path)

    return {"mse": float(mse), "mae": float(mae)}, {"r2": float(r2)}

def get_prediction_for_day_hour(matrix, day, hour):
    """
    Get the prediction for a specific day and hour from the weekly predictions matrix.
    
    Parameters:
    - matrix (numpy array): The result matrix of shape (168, 1) with weekly predictions.
    - day (int): Day of the week (0=Monday, 6=Sunday).
    - hour (int): Hour of the day (0 to 23).
    
    Returns:
    - float: The prediction for the specified day and hour.
    """
    # Calculate the index
    index = day * 24 + hour
    
    # Ensure the index is within the valid range
    if index < 0 or index >= matrix.shape[0]:
        raise ValueError("Invalid day or hour value.")
    
    # Access the prediction
    return matrix[index, 0]

def predictOne(day, hour):
    global model
    global scaler
    global x_test
    global y_test
    
    result_scaler = model.predict(x_test)

    result = scaler.inverse_transform(result_scaler)
    prediction = get_prediction_for_day_hour(result,day,hour)
    
    log.info(prediction)
    return {"prediction": str(prediction)}


def plot_graphs_fn(before_after, file_name, global_date, df_result, today, today_path):
    plt.figure(figsize=(20, 10))
    log.info(len(df_result["result"]))
    plt.plot(df_result["car_sum_real"][:500], label="real")
    plt.plot(df_result["result"][:500], label="prediction")
    plt.legend()
    plt.title("1-D Convolutional Neural Network prediction ")
    
   
    df_result["result"].to_csv(f"{today_path}/{today}.csv")

    # Regular expression to match files like "run0.json", "run1.json", etc.
    pattern = re.compile(r'^run(\d+)\.json$')

    # Find all files that match the pattern and extract the number part
    count = -1
    for path in os.listdir(today_path):
        if os.path.isfile(os.path.join(today_path, path)):
            if pattern.match(path):
                count += 1

    count = 0 if count == -1 else count

    runi = f"run{count}"

    plt.savefig(today_path + "/predict_" + str(before_after) + "_agg_model_" + file_name + "_" + runi + ".png")
