import os
import time
import re
from datetime import timedelta
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
from pandas import DataFrame
import numpy as np
from math import sqrt
from sklearn.model_selection import train_test_split
import tensorflow as tf
from keras.callbacks import EarlyStopping
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, mean_absolute_percentage_error , explained_variance_score
from tensorflow.keras.preprocessing.sequence import pad_sequences
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV, train_test_split
from tensorflow.keras.layers import Dense, LSTM, Masking, Dropout
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Sequential

from common import logger
from common import containers_info

global dataset
dataset = os.environ.get("DATASET")
logging_level = logger.getLevelName(os.environ.get("LOGGING_LEVEL"))

x_train = []
y_train = []
x_test = []
y_test = []
scaler = None
global_filename = ""
global_date = ""
occupied = 0


# Configure logging
log = logger.setup_logger(containers_info.get_current_container_name(), logging_level)
# 1. Function to create time-based features
def add_time_features(df):
    df['hour'] = df.index.hour
    df['day_of_week'] = df.index.dayofweek
    df['day_of_month'] = df.index.day
    df['month'] = df.index.month
    df['is_weekend'] = df['day_of_week'].apply(lambda x: 1 if x >= 5 else 0)
    return df

# 4. Build the LSTM model
def build_lstm_model(n_steps):
    model = Sequential()
    #model.add(Masking(mask_value=0., input_shape=(n_steps, 1)))  # Assuming 0 is used to pad missing values
    model.add(LSTM(128, activation='relu', input_shape=(n_steps, 1), return_sequences=False))
    #model.add(Dropout(0.05))  # Regularization to avoid overfitting
    model.add(Dense(1))
    optimizer = Adam(learning_rate=0.0001)
    model.compile(optimizer, loss='mse')
    return model

def get_parameters():
    log.info("Getting Parameters!")
    global model
    parameters = model.get_weights()

    return parameters


def set_parameters(parameters):
    global model
    model.set_weights(parameters)
    log.info("Parameters Set!")


def in_usage(value):
    global occupied
    occupied = value
    if value == 1:
        log.info("ML app is now in use, and blocked")
    elif value == 0:
        log.info("ML app is now freed")

def get_data():
    global seq_length
    global scaler
    global global_date
    global global_filename
    ##########################################################################################
    # ASK TO ANOTHER ENTITY WICH IS THE DATASET FILE AND OTHER RESTRICTIONS
    file_name = dataset
    date = "2024-05-16"
    ##########################################################################################

   
    global_filename = str(file_name).split(".")[0]
    
    global_date = str(date)

    df = pd.read_csv("/app/dataset/" + str(file_name), index_col="timestamp", parse_dates=True)
    
    day = pd.to_datetime(str(date))
    #df = df[day - timedelta(days=28) :]

    # Preprocessing
    
    # Adding time features
    df = add_time_features(df)
    
   
    # Scale data (excluding time features)
    scaler = MinMaxScaler(feature_range=(0, 2))
    scaled_data = scaler.fit_transform(df[['car_sum']])
    
    # Create a new DataFrame with scaled 'car_sum' and time features
    df_scaled = pd.DataFrame(scaled_data, index=df.index, columns=['car_sum'])
    #df_scaled = pd.concat([df_scaled, df[['hour', 'day_of_week', 'day_of_month', 'month', 'is_weekend']]], axis=1)

    # Create sequences and labels for training

    log.info(df_scaled.head())
    log.info(len(df_scaled))
    
    seq_length = int(0.25 * len(df_scaled))
    X, y = [], []
    for i in range(len(df_scaled) - seq_length):
        X.append(df_scaled[i:i + seq_length])
      
        
            # log.info(df_scaled)
            # log.info(df_scaled[i])
        y.append(df_scaled.iloc[i + seq_length])
        log.info(f"X: {X[-1]}")  # Log the most recent sequence in X
        log.info(f"y: {y[-1]}")  # Log the most recent sequence in y
       


    X, y = np.array(X), np.array(y)

    # Split the data into training and test sets
    global train_size
    train_size = int(0.8 * len(X))
    global x_train
    global y_train
    global x_test
    global y_test
   

   # Convert X and y to numpy arrays
    X, y = np.array(X), np.array(y)

    # Log the shapes of the full sequence arrays
    log.info(f"Full X shape: {X.shape}")
    log.info(f"Full y shape: {y.shape}")

    log.info(f"Full X: {X}")
    log.info(f"Full y: {y}")

    # Now split the data into training and test sets using train_test_split
    x_train, x_test, y_train, y_test = train_test_split(X, y, test_size=0.2, shuffle=False)


    # Log the shapes of the train and test sets
    log.info(f"x_train shape: {x_train.shape}")
    log.info(f"y_train shape: {y_train.shape}")
    log.info(f"x_test shape: {x_test.shape}")
    log.info(f"y_test shape: {y_test.shape}")
    log.info("Get data successfully!")

    return len(x_train)

##############################################################################
##############################################################################
##############################################################################
def fit(config):
    log.info("----------------  FIT  ----------------- ")
    log.debug("config: " + str(config))
    global scaler
    global y_test
    global x_test
    global x_train
    global model
    global today
    global today_path
    log.info(type(x_train))
    log.info(x_train.shape)
    input_size = x_train.shape[2]
    n_steps = 24
    hidden_size = 128
    output_size = 1
    learning_rate = 0.001
    num_epochs=int(config["epochs"]),
    batch_size=int(config["batch_size"]),
    validation_data=(x_test, y_test),

 

    # Assuming x_train is of shape (n_samples, 47, n_features) and x_test is of shape (n_samples, 12, n_features)
    x_test_padded = pad_sequences(x_test, maxlen=len(x_train), padding='post', dtype='float32', value=0)

    # Ensure y_test is also adjusted accordingly, though in most cases it wouldn't require padding if it's just labels.


    # Initialize the model, loss function, and optimizer
    # Build LSTM model
    model = build_lstm_model(n_steps)
    # Train the model
    # x_train_features = x_train[['car_sum']]  # Select the feature(s) correctly
    # y_train_targets = y_train[['car_sum']]    # Ensure you have the correct target(s)

    # x_test_features = x_test_padded[['car_sum']]
    # y_test_targets = y_test[['car_sum']]

    # Define EarlyStopping callback
    early_stopping = EarlyStopping(
        monitor='val_loss',      # Monitor validation loss
        patience=10,             # Number of epochs to wait for improvement
        restore_best_weights=True  # Restore the best weights after stopping
    )

    # Fit the model with early stopping
    history = model.fit(
        x_train, 
        y_train, 
        epochs=32, 
        validation_data=(x_test, y_test), 
        verbose=0,
        callbacks=[early_stopping]  # Add the callback here
    )

    plt.plot(history.history['loss'], label='Train loss')
    plt.plot(history.history['val_loss'], label='Test loss')
    plt.legend()
    today = datetime.now()
    today_path = f"/app/logs/LSTM_{dataset}_{today}"
    if not os.path.exists(today_path):
            os.makedirs(today_path)

    plt.savefig(today_path + "/history.png")
   

    return len(x_test), {"input_size": input_size, "hidden_size": hidden_size, "input_size": output_size, "learning_rate":learning_rate,
                        "num_epochs":num_epochs, "batch_size": batch_size}

def evaluate(config):
    log.info("----------------  EVALUATE  ----------------- ")
    log.debug("config: " + str(config))

    global model
    global x_test
    global y_test

    start_time_evaluate = time.time()
    loss, r_square = model.evaluate(x_test, y_test, steps=int(config["val_steps"]))
    log.info("Evaluate Duration: " + str((time.time() - start_time_evaluate)))

    return float(loss), len(x_test), {"r2_score": float(r_square)}

def predict(plot_graphs, before_after):
    global model
    global y_test
    global today
    global today_path
    
  # Make predictions
    predictions = model.predict(x_test)
    predictions = scaler.inverse_transform(predictions)
    #y_test_inverse = scaler.inverse_transform(y_test.reshape(-1, 1))
    log.info(y_test.shape)
    log.info(y_test)
    #log.info(y_test_inverse.shape)
    log.info(predictions.shape)
    log.info(predictions)
    car_sum_index = 1 
    # Ensure y_test is reshaped to 2D if it is currently 1D
    y_test_inverse = scaler.inverse_transform(y_test.reshape(-1, 1))

    log.info(f'y_test_inverse shape: {y_test_inverse.shape}')
    log.info(f'y_test_inverse: {y_test_inverse}')

    log.info(f'predictions shape: {predictions.shape}')
    log.info(f'predictions: {predictions}')
    # Extract the predictions for car_sum
    #predictions_car_sum = predictions[:, :, car_sum_index]  # Extracting the car_sum dimension


    # Calculate metrics
    mse = mean_squared_error(y_test_inverse, predictions)
    rmse = np.sqrt(mse)
    mae = mean_absolute_error(y_test_inverse, predictions)
    r2 = r2_score(y_test_inverse, predictions)
    explained_variance = explained_variance_score(y_test_inverse, predictions)

    
    # Plot results
    if plot_graphs:
        LSTMplot_graphs_fn(global_filename, global_date, y_test_inverse, predictions, today, today_path)
    
    # Save predictions

     # Define the log file path
    log_file = os.path.join(today_path, 'metrics_log.txt')

    # Write metrics to the file
    with open(log_file, 'a') as f:
        f.write(f"Dataset: {global_filename}")
        f.write(f"Evaluation MSE: {mse:.3f}\n")
        f.write(f"Evaluation RMSE: {rmse:.3f}\n")
        f.write(f"Evaluation MAE: {mae:.3f}\n")
        f.write(f"Evaluation R2_SCORE: {r2:.3f}\n")
        f.write(f"Evaluation Explained Variance: {explained_variance:.3f}\n")
        f.write("\n")  # Add a newline for separation between log entries
  
    y_pred_df = pd.DataFrame(predictions )
    y_pred_df.to_csv(f"{today_path}/{today}.csv", index=False)
    return {"mse": float(mse), "mae": float(mae)}, {"r2": float(r2)}

def get_prediction_for_day_hour(matrix, day, hour):
    """
    Get the prediction for a specific day and hour from the weekly predictions matrix.
    
    Parameters:
    - matrix (numpy array): The result matrix of shape (168, 1) with weekly predictions.
    - day (int): Day of the week (0=Monday, 6=Sunday).
    - hour (int): Hour of the day (0 to 23).
    
    Returns:
    - float: The prediction for the specified day and hour.
    """
    # Calculate the index
    index = day * 24 + hour
    
    # Ensure the index is within the valid range
    if index < 0 or index >= matrix.shape[0]:
        raise ValueError("Invalid day or hour value.")
    
    # Access the prediction
    return matrix[index, 0]

def predictOne(day, hour):
    with torch.no_grad():
        #log.info(type(x_test))
        X_test_tensor = torch.tensor(x_test, dtype=torch.float32)
        #log.info(type(X_test_tensor))
        #log.info(X_test_tensor)
        y_pred = model(X_test_tensor).numpy()
        #log.info(type(y_pred))
        #log.info(type(scaler))
        #log.info(y_pred)
        y_pred = scaler.inverse_transform(y_pred)

    
    y_pred_df = pd.DataFrame(y_pred)
    prediction = get_prediction_for_day_hour(y_pred_df,day,hour)
    
    log.info(prediction)
    return {"prediction": str(prediction)}

def LSTMplot_graphs_fn(file_name, global_date, y_test, y_pred, today, today_path):
        global train_size
        global seq_length
        plt.figure(figsize=(20, 10))
        plt.plot(y_test, label='Actual')
        plt.plot(y_pred, label='Predicted')
        plt.legend()
        plt.title("Long-Short Term Memory prediction")
        
        

        # Regular expression to match files like "run0.json", "run1.json", etc.
        pattern = re.compile(r'^run(\d+)\.json$')

        # Find all files that match the pattern and extract the number part
        count = -1
        for path in os.listdir(today_path):
            if os.path.isfile(os.path.join(today_path, path)):
                if pattern.match(path):
                    count += 1

        count = 0 if count == -1 else count

        runi = f"run{count}"

        plt.savefig(today_path + "/predict_" + file_name + "_" + runi + ".png")
        return today, today_path
