import os
import time
import re
from datetime import timedelta
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
from pandas import DataFrame
import numpy as np
from math import sqrt
import tensorflow as tf
from keras.callbacks import EarlyStopping
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, mean_absolute_percentage_error, explained_variance_score
from sklearn.preprocessing import MinMaxScaler
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV, train_test_split
from tensorflow.keras.layers import Dense, Dropout, SimpleRNN
from tensorflow.keras.optimizers import Adam
from sklearn.model_selection import KFold

from common import logger
from common import containers_info

global dataset
dataset = os.environ.get("DATASET")
logging_level = logger.getLevelName(os.environ.get("LOGGING_LEVEL"))

x_train = []
y_train = []
x_test = []
y_test = []
scaler = None
global_filename = ""
global_date = ""
occupied = 0


# Configure logging
log = logger.setup_logger(containers_info.get_current_container_name(), logging_level)


# # [] -> actual parameters, 0 -> value of the metric
# parameters_best = ([],0)
# parameters_before_last_agg = ([], 0)


#Random Forest model
modelRF = RandomForestRegressor(n_estimators=100, random_state=0)

def get_parameters():
    log.info("Getting Parameters!")
    global model
    parameters = model.get_weights()

    return parameters


def set_parameters(parameters):
    global model
    model.set_weights(parameters)
    log.info("Parameters Set!")


def in_usage(value):
    global occupied
    occupied = value
    if value == 1:
        log.info("ML app is now in use, and blocked")
    elif value == 0:
        log.info("ML app is now freed")

def split_data(df,split_index):
    global x_train, y_train, x_test, y_test
    
    # Validate the split index
    if split_index >= len(df):
        log.info(split_index)
        log.info(len(df))
        raise ValueError(f"Split index {split_index} is greater than or equal to the length of the DataFrame {len(df)}.")

    split_index = int(len(df)*0.75)
    
    # Select training data
    x_train = df.iloc[:split_index][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_train = df.iloc[:split_index][["car_sum_scaler"]]
    
    # Select testing data
    x_test = df.iloc[split_index:][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_test = df.iloc[split_index:][["car_sum_scaler"]]
    
    # Print the shapes of the resulting datasets (optional)
    print("Shapes:")
    print("x_train:", x_train.shape)
    print("y_train:", y_train.shape)
    print("x_test:", x_test.shape)
    print("y_test:", y_test.shape)

def get_data():
    ##########################################################################################
    # ASK TO ANOTHER ENTITY WICH IS THE DATASET FILE AND OTHER RESTRICTIONS
    file_name = dataset
    date = "2022-05-16"
    ##########################################################################################

    global global_filename
    global_filename = str(file_name).split(".")[0]
    global global_date
    global_date = str(date)
    global df
    df = pd.read_csv("/app/dataset/" + str(file_name), index_col="timestamp", parse_dates=True)
    day = pd.to_datetime(str(date))
    log.info("df")
    log.info(len(df))
    log.info(df)
    #df = df.resample('H').mean()  # For hourly data
    #df = df.interpolate(method='linear')  # Fill missing values if needed

    #df = df[day - timedelta(days=28) :]

    start_dataset = df.index[0]
    end_dataset = df.index[-1]
    log.info("start timestamp: " + str(start_dataset))
    log.info("end timestamp: " + str(end_dataset))

    global scaler
    scaler = MinMaxScaler()
    scaler.fit(df["car_sum"].to_numpy().reshape(-1, 1))
    df["car_sum_scaler"] = scaler.transform(df["car_sum"].to_numpy().reshape(-1, 1))

        # Calculate the split index
    split_index = int(len(df)*0.23)  # Example: Split after 4 days (24*4 hours)

    df["car_sumB"] = df["car_sum"].shift((split_index))
    df["car_sumBB"] = df["car_sum"].shift((split_index * 2))
    df["car_sumBBB"] = df["car_sum"].shift((split_index * 3))
   
    df["car_sumB_scaler"] = df["car_sum_scaler"].shift((split_index))
    df["car_sumBB_scaler"] = df["car_sum_scaler"].shift((split_index * 2))
    df["car_sumBBB_scaler"] = df["car_sum_scaler"].shift((split_index * 3))
    
    df["car_sum_diff"] = df["car_sum"].diff()  # Calculate difference between consecutive values

    df.fillna(0, inplace=True)  # Example

    # Call the function to split the data
    split_data(df, split_index)

    # parameters_current = get_parameters()
    # metric_value_current = evaluate()[2]
    # global parameters_before_last_agg
    # global parameters_best
    # parameters_before_last_agg = (parameters_current, metric_value_current)
    # parameters_best = (parameters_current, metric_value_current)

    log.info("Get data successfully!")

    return len(x_train)

def fit(config):
    log.info("----------------  FITRF  ----------------- ")
    log.debug("config: " + str(config))

    global modelRF
    global x_train
    global y_train
    global x_test
    global y_test

    #es = EarlyStopping(monitor="val_loss", mode="min", patience=5)

   
    # Set up K-Fold cross-validation
    kf = KFold(n_splits=5, shuffle=True, random_state=0)  # 5 folds
    fold_results = []  # Store results for each fold

    start_time_fit = time.time()
    
    # K-Fold cross-validation
    for fold, (train_index, val_index) in enumerate(kf.split(x_train)):
        log.info(f"Training fold {fold + 1}")
        
        # Split data into training and validation sets
        x_train_fold, x_val_fold = x_train.iloc[train_index], x_train.iloc[val_index]
        y_train_fold, y_val_fold = y_train.iloc[train_index], y_train.iloc[val_index]

        # Fit the model on the current fold
        modelRF.fit(x_train_fold, y_train_fold)

        # Evaluate the model on the validation set
        val_predictions = modelRF.predict(x_val_fold)

        # Calculate metrics
        val_mse = mean_squared_error(y_val_fold, val_predictions)
        fold_results.append(val_mse)
        log.info(f"Fold {fold + 1} MSE: {val_mse}")

    # Calculate the average MSE across all folds
    avg_mse = np.mean(fold_results)
    log.info(f"Average MSE across folds: {avg_mse}")

    log.info("Fit Duration: " + str((time.time() - start_time_fit)))
    
    # Extract feature importances and out-of-bag score if applicable
    feature_importances = modelRF.feature_importances_
    oob_score = modelRF.oob_score_ if hasattr(modelRF, 'oob_score_') else None
    log.info(feature_importances)
    log.info(oob_score)

    results = {
        "fit_duration": str((time.time() - start_time_fit)),
        "avg_mse": avg_mse,  # Store average MSE
        "feature_importances": str(feature_importances),
        "oob_score": str(oob_score),
    }
    return len(x_train), results

def evaluate(config):
    log.info("----------------  EVALUATE  ----------------- ")
    log.debug("config: " + str(config))

    global modelRF
    global x_test
    global y_test

    start_time_evaluate = time.time()
    loss, r_square = modelRF.evaluate(x_test, y_test, steps=int(config["val_steps"]))
    log.info("Evaluate Duration: " + str((time.time() - start_time_evaluate)))

    return float(loss), len(x_test), {"r2_score": float(r_square)}

def predict(plot_graphs, before_after):
    global modelRF
    global scaler
    global x_test
    global y_test

    # Train and Test predictions
    y_train_pred = modelRF.predict(x_train)
    y_test_pred = modelRF.predict(x_test)

    # Inverse transform the predictions and actual test values
    result = scaler.inverse_transform(y_test_pred.reshape(-1, 1))
    car_sum_real = scaler.inverse_transform(y_test)
    log.info(f"Result: {result}")
    log.info(f"Real: {car_sum_real}")

    # Calculate metrics for training set
    train_mse = mean_squared_error(y_train, y_train_pred)
    train_rmse = np.sqrt(train_mse)
    train_mae = mean_absolute_error(y_train, y_train_pred)
    train_r2 = r2_score(y_train, y_train_pred)
    train_explained_variance = explained_variance_score(y_train, y_train_pred)
    #train_mape = np.mean(np.abs((y_train - y_train_pred) / y_train)) * 100  # Mean Absolute Percentage Error

  
    test_mse = mean_squared_error(car_sum_real, result)
    test_rmse = mean_squared_error(car_sum_real, result, squared=False)
    test_mae = mean_absolute_error(car_sum_real, result)
    test_r2 = r2_score(car_sum_real, result)
    test_explained_variance = explained_variance_score(car_sum_real, result)
    #est_mape = np.mean(np.abs((y_test - y_test_pred) / y_test)) * 100  # Mean Absolute Percentage Error

    # Collect results in a dictionary
    results = {
        "train_mse": train_mse,
        "train_rmse": train_rmse,
        "train_mae": train_mae,
        "train_r2_score": train_r2,
        "train_explained_variance": train_explained_variance,
        #"train_mape": train_mape,
        "test_mse": test_mse,
        "test_rmse": test_rmse,
        "test_mae": test_mae,
        "test_r2_score": test_r2,
        "test_explained_variance": test_explained_variance,
        #"test_mape": test_mape,
    }
     # Save Logs to a file
    today = datetime.now()
    today_path = f"/app/logs/RF_{dataset}_{today}"
    if not os.path.exists(today_path):
        os.makedirs(today_path)
     # Define the log file path
    log_file = os.path.join(today_path, 'metrics_log.txt')
    log.info("Predict Results: " + str(results))

    # Write metrics to the file
    with open(log_file, 'a') as f:
        f.write(f"Dataset: {global_filename}\n")
        f.write(f"Evaluation MSE: {test_mse:.3f}\n")
        f.write(f"Evaluation RMSE: {test_rmse:.3f}\n")
        f.write(f"Evaluation MAE: {test_mae:.3f}\n")
        #f.write(f"Evaluation MAPE: {test_mape:.3f}\n")
        f.write(f"Evaluation R2_SCORE: {test_r2:.3f}\n")
        f.write(f"Evaluation Explained Variance: {test_explained_variance:.3f}\n")
        f.write("\n")  # Add a newline for separation between log entries

    # Plotting if specified
    if plot_graphs != "false":
        RFplot_graphs_fn(str(global_filename), str(global_date), car_sum_real, result,today, today_path)


    return results, test_r2

def RFplot_graphs_fn(file_name, global_date, car_sum_real, result, today, today_path):
    
    plt.figure(figsize=(10, 10))
    plt.plot(car_sum_real[:500], label="real")
    plt.plot(result[:500], label="prediction")
    plt.legend()
    plt.title("Random Forest prediction  ")
    
   
    y_pred_df = pd.DataFrame(result)
    y_pred_df.to_csv(f"{today_path}/{today}.csv", index=False) 

    # Regular expression to match files like "run0.json", "run1.json", etc.
    pattern = re.compile(r'^run(\d+)\.json$')

    # Find all files that match the pattern and extract the number part
    count = -1
    for path in os.listdir(today_path):
        if os.path.isfile(os.path.join(today_path, path)):
            if pattern.match(path):
                count += 1

    count = 0 if count == -1 else count

    runi = f"run{count}"

    plt.savefig(today_path + "/predict_" + file_name + "_" + runi + ".png")

def get_prediction_for_day_hour(matrix, day, hour):
    """
    Get the prediction for a specific day and hour from the weekly predictions matrix.
    
    Parameters:
    - matrix (numpy array): The result matrix of shape (168, 1) with weekly predictions.
    - day (int): Day of the week (0=Monday, 6=Sunday).
    - hour (int): Hour of the day (0 to 23).
    
    Returns:
    - float: The prediction for the specified day and hour.
    """
    # Calculate the index
    index = day * 24 + hour
    
    # Ensure the index is within the valid range
    if index < 0 or index >= matrix.shape[0]:
        raise ValueError("Invalid day or hour value.")
    
    # Access the prediction
    return matrix[index, 0]

def predictOne(day, hour):
    global model
    global scaler
    global x_test
    global y_test
    
    result_scaler = model.predict(x_test)

    result = scaler.inverse_transform(result_scaler)
    prediction = get_prediction_for_day_hour(result,day,hour)
    
    log.info(prediction)
    return {"prediction": str(prediction)}

