import os
import time
import re
from datetime import timedelta
from datetime import datetime

import matplotlib.pyplot as plt
import pandas as pd
from pandas import DataFrame
import numpy as np
from math import sqrt
import tensorflow as tf
from keras.callbacks import EarlyStopping
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score, mean_absolute_percentage_error, explained_variance_score
from sklearn.preprocessing import MinMaxScaler
from sklearn.ensemble import RandomForestRegressor
from sklearn.model_selection import GridSearchCV, train_test_split
from tensorflow.keras.layers import Dense, Dropout, SimpleRNN
from tensorflow.keras.optimizers import Adam
from sklearn.model_selection import KFold
from xgboost import XGBRegressor
from common import logger
from common import containers_info

global dataset
dataset = os.environ.get("DATASET")
logging_level = logger.getLevelName(os.environ.get("LOGGING_LEVEL"))

x_train = []
y_train = []
x_test = []
y_test = []
scaler = None
global_filename = ""
global_date = ""
occupied = 0


# Configure logging
log = logger.setup_logger(containers_info.get_current_container_name(), logging_level)


# # [] -> actual parameters, 0 -> value of the metric
# parameters_best = ([],0)
# parameters_before_last_agg = ([], 0)


# Initialize XGBoost model
modelXGB = XGBRegressor(n_estimators=100, max_depth=5, learning_rate=0.1, random_state=0)


def get_parameters():
    log.info("Getting Parameters!")
    global model
    parameters = model.get_weights()

    return parameters


def set_parameters(parameters):
    global model
    model.set_weights(parameters)
    log.info("Parameters Set!")


def in_usage(value):
    global occupied
    occupied = value
    if value == 1:
        log.info("ML app is now in use, and blocked")
    elif value == 0:
        log.info("ML app is now freed")

def split_data(df,split_index):
    global x_train, y_train, x_test, y_test
    split_index = int(len(df) * 0.75)
    x_train = df.iloc[:split_index][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_train = df.iloc[:split_index][["car_sum_scaler"]]
    x_test = df.iloc[split_index:][["car_sumB_scaler", "car_sumBB_scaler", "car_sumBBB_scaler"]]
    y_test = df.iloc[split_index:][["car_sum_scaler"]]

def get_data():
    global global_filename
    ##########################################################################################
    # ASK TO ANOTHER ENTITY WICH IS THE DATASET FILE AND OTHER RESTRICTIONS
    file_name = dataset
    global_filename = file_name
    date = "2022-05-16"
    ##########################################################################################

    global df, scaler
    df = pd.read_csv("/app/dataset/" + dataset, index_col="timestamp", parse_dates=True)
    scaler = MinMaxScaler()
    df["car_sum_scaler"] = scaler.fit_transform(df["car_sum"].to_numpy().reshape(-1, 1))
    split_index = int(len(df) * 0.23)
    df["car_sumB_scaler"] = df["car_sum_scaler"].shift((split_index))
    df["car_sumBB_scaler"] = df["car_sum_scaler"].shift((split_index * 2))
    df["car_sumBBB_scaler"] = df["car_sum_scaler"].shift((split_index * 3))
    df.fillna(0, inplace=True)
    split_data(df, split_index)
    return len(x_train)

def fit(config):
    log.info("----------------  FIT --------------- ")
    log.debug("config: " + str(config))

    global modelXGB
    kf = KFold(n_splits=5, shuffle=True, random_state=0)
    fold_results = []
    start_time_fit = time.time()
    
    for fold, (train_index, val_index) in enumerate(kf.split(x_train)):
        x_train_fold, x_val_fold = x_train.iloc[train_index], x_train.iloc[val_index]
        y_train_fold, y_val_fold = y_train.iloc[train_index], y_train.iloc[val_index]
        modelXGB.fit(x_train_fold, y_train_fold)
        val_predictions = modelXGB.predict(x_val_fold)
        val_mse = mean_squared_error(y_val_fold, val_predictions)
        fold_results.append(val_mse)
    
    avg_mse = np.mean(fold_results)
    fit_duration = time.time() - start_time_fit
    results = {
        "fit_duration": fit_duration,
        "avg_mse": avg_mse,
        "feature_importances": modelXGB.feature_importances_.tolist(),
    }
    return len(x_train), results

# Evaluation function
def evaluate_xgb(config):
    start_time_evaluate = time.time()
    predictions = modelXGB.predict(x_test)
    test_mse = mean_squared_error(y_test, predictions)
    test_rmse = np.sqrt(test_mse)
    test_mae = mean_absolute_error(y_test, predictions)
    test_r2 = r2_score(y_test, predictions)
    test_explained_variance = explained_variance_score(y_test, predictions)
    evaluation_duration = time.time() - start_time_evaluate

    results = {
        "test_mse": test_mse,
        "test_rmse": test_rmse,
        "test_mae": test_mae,
        "test_r2_score": test_r2,
        "test_explained_variance": test_explained_variance,
    }
    return results, test_r2


def evaluate(config):
    log.info("----------------  EVALUATE  ----------------- ")
    log.debug("config: " + str(config))

    global modelXGB
    global x_test
    global y_test

    start_time_evaluate = time.time()
    loss, r_square = modelXGB.evaluate(x_test, y_test, steps=int(config["val_steps"]))
    log.info("Evaluate Duration: " + str((time.time() - start_time_evaluate)))

    return float(loss), len(x_test), {"r2_score": float(r_square)}

def predict(plot_graphs, before_after):
    predictions = modelXGB.predict(x_test)
    result = scaler.inverse_transform(predictions.reshape(-1, 1))
    real_values = scaler.inverse_transform(y_test)
    
    
    test_mse= mean_squared_error(real_values, result)
    test_rmse= np.sqrt(mean_squared_error(real_values, result))
    test_mae= mean_absolute_error(real_values, result)
    test_r2_score= r2_score(real_values, result)
    test_explained_variance= explained_variance_score(real_values, result)

    results = {
        "test_mse": mean_squared_error(real_values, result),
        "test_rmse": np.sqrt(mean_squared_error(real_values, result)),
        "test_mae": mean_absolute_error(real_values, result),
        "test_r2_score": r2_score(real_values, result),
        "test_explained_variance": explained_variance_score(real_values, result),
    }
    
    
    # Optional plotting
    if plot_graphs:
        plt.figure(figsize=(10, 10))
        plt.plot(real_values[:500], label="real")
        plt.plot(result[:500], label="prediction")
        plt.legend()
        plt.title("XGBoost Predictions")
        plt.savefig("/app/logs/xgb_predictions.png")
     # Save Logs to a file
    today = datetime.now()
    today_path = f"/app/logs/XGB_{dataset}_{today}"
    if not os.path.exists(today_path):
        os.makedirs(today_path)
     # Define the log file path
    log_file = os.path.join(today_path, 'metrics_log.txt')
    log.info("Predict Results: " + str(results))

    # Write metrics to the file
    with open(log_file, 'a') as f:
        f.write(f"Dataset: {global_filename}\n")
        f.write(f"Evaluation MSE: {test_mse:.3f}\n")
        f.write(f"Evaluation RMSE: {test_rmse:.3f}\n")
        f.write(f"Evaluation MAE: {test_mae:.3f}\n")
        #f.write(f"Evaluation MAPE: {test_mape:.3f}\n")
        f.write(f"Evaluation R2_SCORE: {test_r2_score:.3f}\n")
        f.write(f"Evaluation Explained Variance: {test_explained_variance:.3f}\n")
        f.write("\n")  # Add a newline for separation between log entries

    # Plotting if specified
    if plot_graphs != "false":
        RFplot_graphs_fn(str(global_filename), str(global_date), real_values, result,today, today_path)


    return results, test_r2_score

def RFplot_graphs_fn(file_name, global_date, car_sum_real, result, today, today_path):
    
    plt.figure(figsize=(10, 10))
    plt.plot(car_sum_real[:500], label="real")
    plt.plot(result[:500], label="prediction")
    plt.legend()
    plt.title("XGBoost prediction")
    
   
    y_pred_df = pd.DataFrame(result)
    y_pred_df.to_csv(f"{today_path}/{today}.csv", index=False) 

    # Regular expression to match files like "run0.json", "run1.json", etc.
    pattern = re.compile(r'^run(\d+)\.json$')

    # Find all files that match the pattern and extract the number part
    count = -1
    for path in os.listdir(today_path):
        if os.path.isfile(os.path.join(today_path, path)):
            if pattern.match(path):
                count += 1

    count = 0 if count == -1 else count

    runi = f"run{count}"

    plt.savefig(today_path + "/predict_" + file_name + "_" + runi + ".png")

def get_prediction_for_day_hour(matrix, day, hour):
    """
    Get the prediction for a specific day and hour from the weekly predictions matrix.
    
    Parameters:
    - matrix (numpy array): The result matrix of shape (168, 1) with weekly predictions.
    - day (int): Day of the week (0=Monday, 6=Sunday).
    - hour (int): Hour of the day (0 to 23).
    
    Returns:
    - float: The prediction for the specified day and hour.
    """
    # Calculate the index
    index = day * 24 + hour
    
    # Ensure the index is within the valid range
    if index < 0 or index >= matrix.shape[0]:
        raise ValueError("Invalid day or hour value.")
    
    # Access the prediction
    return matrix[index, 0]

def predictOne(day, hour):
    global model
    global scaler
    global x_test
    global y_test
    
    result_scaler = model.predict(x_test)

    result = scaler.inverse_transform(result_scaler)
    prediction = get_prediction_for_day_hour(result,day,hour)
    
    log.info(prediction)
    return {"prediction": str(prediction)}

