import json
import docker
import os
import time
import logging
from datetime import datetime,date
import hashlib
import requests
import psutil
import socket
import tarfile
import io
import aiohttp
import asyncio

async def async_call_api_endpoint(endpoint, ip_addr, ports):

    timeout_value = aiohttp.ClientTimeout(sock_connect=10, sock_read=30)
    method = endpoint.get("method", "GET").upper()
    url = f"http://{ip_addr}:{ports}{endpoint.get('url')}"
    headers = endpoint.get("headers", {})
    body = endpoint.get("body", {})

    logging.info(f"Calling API endpoint: {method} {url} {body}")
   

    try:
        async with aiohttp.ClientSession(timeout=timeout_value) as session:
            await asyncio.sleep(45)
            if method == "GET":
                logging.info("Inside AsyncRequest")
                async with session.get(url, headers=headers) as response:
                   
                    await handle_response(url, response)
                logging.info("Left AsyncRequest")
            elif method == "POST":
                async with session.post(url, headers=headers, json=body) as response:
                   
                    await handle_response(url, response)
            else:
                logging.warning(f"Unsupported HTTP method: {method}")
    except Exception as e:
        logging.error(f"Error calling API endpoint: {e}")


async def handle_response(url, response):
    """Handle and log the API response."""
    if response.status == 200:
        logging.info(f"API call to {url} successful: {response.status}")
    else:
        text = await response.text()
        logging.warning(f"API call to {url} failed: {response.status} {text}")

async def async_call_api_endpoints(api_endpoints, ip_addr, ports):
 
    for endpoint in api_endpoints:
        await async_call_api_endpoint(endpoint, ip_addr, ports)  # Wait for each call to complete




ports = 5001 
def get_wireless_ipv4():
    wireless_ipv4 = None
    # Iterate over all network interfaces
    for interface, addrs in psutil.net_if_addrs().items():
        # Check if the interface name suggests a wireless connection
        if interface.startswith('wl'):
            for addr in addrs:
                if addr.family == socket.AF_INET:
                    wireless_ipv4 = addr.address
                    break
            if wireless_ipv4:
                break
    return wireless_ipv4

class FileModifiedException(Exception):
    pass

# Set up logging
logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s - %(levelname)s - %(message)s',
                    handlers=[logging.StreamHandler()])


def calculate_file_hash(file_path):
    with open(file_path, "rb") as f:
        file_hash = hashlib.sha256(f.read()).hexdigest()
    return file_hash


def read_json_file(file_path):
    try:
        with open(file_path, 'r') as file:
            config = json.load(file)
            logging.info(f"Configuration loaded: {config}")
            return config
    except Exception as e:
        logging.info(f"Error reading JSON file: {e}")
        return None

def substitute_env_variables(config):
    for key, value in config.items():
        if isinstance(value, str):
            config[key] = os.path.expandvars(value)
        elif isinstance(value, dict):
            config[key] = {k: os.path.expandvars(v) if isinstance(v, str) else v for k, v in value.items()}
        elif isinstance(value, list):
            config[key] = [os.path.expandvars(v) if isinstance(v, str) else v for v in value]
    return config

def generate_unique_container_name(base_name):
    timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
    unique_name = f"{base_name}-{timestamp}"
    return unique_name

def read_json_file(file_path):
    try:
        with open(file_path, 'r') as file:
            config = json.load(file)
            logging.info(f"Configuration loaded: {config}")
            return config
    except Exception as e:
        logging.info(f"Error reading JSON file: {e}")
        return None

def substitute_env_variables(config):
    # Substitute environment variables in the configuration
    for key, value in config.items():
        if isinstance(value, str):
            config[key] = os.path.expandvars(value)
        elif isinstance(value, dict):
            config[key] = {k: os.path.expandvars(v) if isinstance(v, str) else v for k, v in value.items()}
        elif isinstance(value, list):
            config[key] = [os.path.expandvars(v) if isinstance(v, str) else v for v in value]
    return config

def create_container(config):
    try:
        client = docker.from_env()
        logging.info("Docker client initialized")

        global ports
        image = config.get("image")
        container_name = config.get("name")
        #server_port = config.get("ports")
        volumes = config.get("volumes")
        environment = config.get("environment")
        labels = config.get("labels")
        

        container_id = os.environ["HOSTNAME"]
        container_main = client.containers.get(container_id)
        container_main_name = container_main.name
        parent_container = client.containers.get(container_main_name)
        
        parent_network = parent_container.attrs["HostConfig"]["NetworkMode"]
        logging.info(parent_network)

        logging.info(f"Creating container with image: {image}, name: {container_name}, ports: {ports}, volumes: {volumes}, environment: {environment}, labels: {labels}")

        container = client.containers.run(
            image,
            name=container_name,
            ports={"5000/tcp": ("0.0.0.0", int(ports))},
            volumes=volumes,
            environment=environment,
            labels=labels,
            detach=True,  # Run container in the background
            network_mode=parent_network,
        )
        logging.info(f"Container {container_name} created. ID: {container.id}")
        return container
    except docker.errors.APIError as e:
        logging.info(f"APIError creating container: {e.explanation}")
    except Exception as e:
        logging.info(f"Unexpected error: {e}")

def call_api_endpoints(api_endpoints):
    timeout_value = (3, 50) 
    for endpoint in api_endpoints:
        try:
            ip_addr = "192.168.21.255"#str(get_wireless_ipv4())
            global ports
            method = endpoint.get("method", "GET").upper()
            url = "http://"+ip_addr+":"+str(ports)+endpoint.get("url")
            headers = endpoint.get("headers", {})
            body = endpoint.get("body", {})
            
            logging.info(f"Calling API endpoint: {method} {url}")
            logging.info(method == "GET")
            if method == "GET":
                logging.info("Here")
                response = requests.get(url, headers=headers, timeout=timeout_value)
                
                time.sleep(2)
            elif method == "POST":
                response = requests.post(url, headers=headers, json=body)
                time.sleep(30)
            else:
                logging.warning(f"Unsupported HTTP method: {method}")
                continue

            logging.info(response)
            if response.status_code == 200:
                logging.info(f"API call to {url} successful: {response.status_code}")
            else:
                logging.warning(f"API call to {url} failed: {response.status_code} {response.text}")

        except Exception as e:
            logging.error(f"Error calling API endpoint: {e}")
        time.sleep(1)

def copy_file_from_container(container, src_path, dst_path):
    try:
        logging.info(f"Copying directory from container {container.name} at {src_path} to host at {dst_path}")
        
        # Get the tar archive of the directory from the container
        bits, stat = container.get_archive(src_path)
        
        # Create the destination directory if it doesn't exist
        os.makedirs(dst_path, exist_ok=True)
        
        # Create a tarfile object from the stream of bits
        tar_stream = b''.join(bits)
        tar_file = tarfile.open(fileobj=io.BytesIO(tar_stream))
        
        # Extract the contents of the tarball into the destination directory
        tar_file.extractall(path=dst_path)
        tar_file.close()
        
        logging.info(f"Directory copied successfully from container {container.name} to host at {dst_path}")
    except Exception as e:
        logging.error(f"Error copying directory from container: {e}")

def main():
    file_path = '/usr/src/app/config.json'
    output_path = '/usr/src/app/output/'
    logging.info(f"Monitoring {file_path} for changes...")
    global ports

    try:
        last_hash = calculate_file_hash(file_path)
        while True:
            current_hash = calculate_file_hash(file_path)
            if current_hash != last_hash:
                logging.info("File has changed!")
                last_hash = current_hash
                config = read_json_file(file_path)
                config = substitute_env_variables(config)
                config['name'] = generate_unique_container_name(config.get('name'))
                try:
                    container = create_container(config)
                    if container:
                        asyncio.run(async_call_api_endpoints(
                            config.get('api_endpoints', []),
                            ip_addr="192.168.21.255",  # Use local IP or hostname
                            ports=ports
                        ))

                    ports += 1
                except Exception as e:
                    logging.error(f"Error in main: {e}")
            time.sleep(1)
    except KeyboardInterrupt:
        logging.info("INTERRUPT")
      

if __name__ == "__main__":
    main()
