from urllib.parse import quote, unquote
import base64
import numpy as np
import json

def decode_parameters(parameters_url):
    # Convert the decoded string back to a list of NumPy arrays
    decoded_string = unquote(parameters_url)

    parameters_np = []
    for s in decoded_string.split(";"):
        arr_str, shape_str = s.rsplit("(", 1)
        arr_bytes = base64.b64decode(arr_str)
        arr = np.frombuffer(arr_bytes, dtype=np.float32)
        if shape_str:
            shape = tuple(map(int, shape_str.strip().replace("(", "").replace(")", "").split(",")))
            arr = np.reshape(arr, shape)
        parameters_np.append(arr)

    return parameters_np


def encode_parameters(parameters_np):
    parameters_string = ";".join(
        [
            base64.b64encode(arr.tobytes()).decode() + (f"({','.join(map(str, arr.shape))})" if arr.ndim > 0 else "")
            for arr in parameters_np
        ]
    )
    parameters_encoded_string = quote(parameters_string)

    return parameters_encoded_string


def decode_config(config_url):
    # Decode URL-safe string using URL decoding
    decoded_str = unquote(config_url)
    # Convert decoded string to JSON object
    config_json = json.loads(decoded_str)

    return config_json
